@@ -57,19 +57,25 @@
 """
 
 
-import io
+try:
+    from asyncio import get_running_loop
+except ImportError:  # pragma: no cover
+    from asyncio import get_event_loop as get_running_loop  # Python 3.6 compatibility
+
 import logging
 import os
 import socket
 import struct
 import time
 
+import aiofiles
+
 from . import constants
 from . import exceptions
 from .adb_message import AdbMessage, checksum, unpack
 from .transport.base_transport_async import BaseTransportAsync
 from .transport.tcp_transport_async import TcpTransportAsync
-from .hidden_helpers import FILE_TYPES, DeviceFile, _AdbTransactionInfo, _FileSyncTransactionInfo, _open
+from .hidden_helpers import DeviceFile, _AdbTransactionInfo, _FileSyncTransactionInfo, get_files_to_push
 
 
 _LOGGER = logging.getLogger(__name__)
@@ -97,6 +103,8 @@ class AdbDeviceAsync(object):
         Whether an ADB connection to the device has been established
     _banner : bytearray, bytes
         The hostname of the machine where the Python interpreter is currently running
+    _maxdata: int
+        Maximum amount of data in an ADB packet
     _transport : BaseTransportAsync
         The transport that is used to connect to the device; must be a subclass of :class:`~adb_shell.transport.base_transport_async.BaseTransportAsync`
 
@@ -121,6 +129,19 @@ class AdbDeviceAsync(object):
         self._transport = transport
 
         self._available = False
+        self._maxdata = constants.MAX_PUSH_DATA
+
+    @property
+    def max_chunk_size(self):
+        """Maximum chunk size for filesync operations
+
+        Returns
+        -------
+        int
+            Minimum value based on :const:`adb_shell.constants.MAX_CHUNK_SIZE` and ``_max_data / 2``, fallback to legacy :const:`adb_shell.constants.MAX_PUSH_DATA`
+
+        """
+        return min(constants.MAX_CHUNK_SIZE, self._maxdata // 2) or constants.MAX_PUSH_DATA
 
     @property
     def available(self):
@@ -154,7 +175,7 @@ class AdbDeviceAsync(object):
             1. If the last ``arg0`` was not :const:`adb_shell.constants.AUTH_TOKEN`, raise an exception
             2. Sign the last ``banner`` and send it in an ``b'AUTH'`` message
             3. Unpack the ``cmd``, ``arg0``, and ``banner`` fields from the response via :func:`adb_shell.adb_message.unpack`
-            4. If ``cmd`` is ``b'CNXN'``, return ``banner``
+            4. If ``cmd`` is ``b'CNXN'``, set transfer maxdata size and return ``True``
 
         7. None of the keys worked, so send ``rsa_keys[0]``'s public key; if the response does not time out, we must have connected successfully
 
@@ -201,6 +222,7 @@ class AdbDeviceAsync(object):
 
         # 4. If ``cmd`` is not ``b'AUTH'``, then authentication is not necesary and so we are done
         if cmd != constants.AUTH:
+            self._maxdata = arg1  # CNXN maxdata
             self._available = True
             return True  # return banner
 
@@ -222,10 +244,11 @@ class AdbDeviceAsync(object):
             await self._send(msg, adb_info)
 
             # 6.3. Unpack the ``cmd``, ``arg0``, and ``banner`` fields from the response via :func:`adb_shell.adb_message.unpack`
-            cmd, arg0, _, banner = await self._read([constants.CNXN, constants.AUTH], adb_info)
+            cmd, arg0, arg1, banner = await self._read([constants.CNXN, constants.AUTH], adb_info)
 
-            # 6.4. If ``cmd`` is ``b'CNXN'``, return ``banner``
+            # 6.4. If ``cmd`` is ``b'CNXN'``, we are done
             if cmd == constants.CNXN:
+                self._maxdata = arg1  # CNXN maxdata
                 self._available = True
                 return True  # return banner
 
@@ -241,7 +264,8 @@ class AdbDeviceAsync(object):
         await self._send(msg, adb_info)
 
         adb_info.transport_timeout_s = auth_timeout_s
-        cmd, arg0, _, banner = await self._read([constants.CNXN], adb_info)
+        cmd, arg0, arg1, banner = await self._read([constants.CNXN], adb_info)
+        self._maxdata = arg1  # CNXN maxdata
         self._available = True
         return True  # return banner
 
@@ -415,7 +439,7 @@ class AdbDeviceAsync(object):
             raise exceptions.AdbConnectionError("ADB command not sent because a connection to the device has not been established.  (Did you call `AdbDeviceAsync.connect()`?)")
 
         adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
-        filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_LIST_FORMAT)
+        filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_LIST_FORMAT, maxdata=self._maxdata)
         await self._open(b'sync:', adb_info)
 
         await self._filesync_send(constants.LIST, adb_info, filesync_info, data=device_path)
@@ -432,71 +456,45 @@ class AdbDeviceAsync(object):
 
         return files
 
-    async def pull(self, device_filename, dest_file=None, progress_callback=None, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S):
+    async def pull(self, device_path, local_path, progress_callback=None, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S):
         """Pull a file from the device.
 
         Parameters
         ----------
-        device_filename : str
-            Filename on the device to pull.
-        dest_file : str, file, io.IOBase, None
-            If set, a filename or writable file-like object.
+        device_path : str
+            The file on the device that will be pulled
+        local_path : str
+            The path to where the file will be downloaded
         progress_callback : function, None
-            Callback method that accepts filename, bytes_written and total_bytes, total_bytes will be -1 for file-like
-            objects
+            Callback method that accepts ``device_path``, ``bytes_written``, and ``total_bytes``
         transport_timeout_s : float, None
             Expected timeout for any part of the pull.
         read_timeout_s : float
-            The total time in seconds to wait for a ``b'CLSE'`` or ``b'OKAY'`` command in :meth:`AdbDeviceAsync._read`
-
-        Returns
-        -------
-        bytes, bool
-            The file data if ``dest_file`` is not set. Otherwise, ``True`` if the destination file exists
-
-        Raises
-        ------
-        ValueError
-            If ``dest_file`` is of unknown type.
+            The total time in seconds to wait for a ``b'CLSE'`` or ``b'OKAY'`` command in :meth:`AdbDevice._read`
 
         """
         if not self.available:
             raise exceptions.AdbConnectionError("ADB command not sent because a connection to the device has not been established.  (Did you call `AdbDeviceAsync.connect()`?)")
 
-        if not dest_file:
-            dest_file = io.BytesIO()
-
-        if not isinstance(dest_file, FILE_TYPES + (str,)):
-            raise ValueError("dest_file is of unknown type")
-
         adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
-        filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_PULL_FORMAT)
+        filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_PULL_FORMAT, maxdata=self._maxdata)
 
-        with _open(dest_file, 'wb') as dest:
+        async with aiofiles.open(local_path, 'wb') as stream:
             await self._open(b'sync:', adb_info)
-            await self._pull(device_filename, dest, progress_callback, adb_info, filesync_info)
+            await self._pull(device_path, stream, progress_callback, adb_info, filesync_info)
             await self._close(adb_info)
 
-            if isinstance(dest, io.BytesIO):
-                return dest.getvalue()
-
-            if hasattr(dest, 'name'):
-                return os.path.exists(dest.name)
-
-            # We don't know what the path is, so we just assume it exists.
-            return True
-
-    async def _pull(self, filename, dest, progress_callback, adb_info, filesync_info):
-        """Pull a file from the device into the file-like ``dest_file``.
+    async def _pull(self, device_path, stream, progress_callback, adb_info, filesync_info):
+        """Pull a file from the device into the file-like ``local_path``.
 
         Parameters
         ----------
-        filename : str
-            The file to be pulled
-        dest : _io.BytesIO
+        device_path : str
+            The file on the device that will be pulled
+        stream : AsyncBufferedIOBase
             File-like object for writing to
         progress_callback : function, None
-            Callback method that accepts ``filename``, ``bytes_written``, and ``total_bytes``
+            Callback method that accepts ``device_path``, ``bytes_written``, and ``total_bytes``
         adb_info : _AdbTransactionInfo
             Info and settings for this ADB transaction
         filesync_info : _FileSyncTransactionInfo
@@ -504,37 +502,36 @@ class AdbDeviceAsync(object):
 
         """
         if progress_callback:
-            total_bytes = await self.stat(filename)[1]
-            progress = self._transport_progress(lambda current: progress_callback(filename, current, total_bytes))
+            total_bytes = await self.stat(device_path)[1]
+            progress = self._transport_progress(lambda current: progress_callback(device_path, current, total_bytes))
             next(progress)
 
-        await self._filesync_send(constants.RECV, adb_info, filesync_info, data=filename)
+        await self._filesync_send(constants.RECV, adb_info, filesync_info, data=device_path)
         async for cmd_id, _, data in self._filesync_read_until([constants.DATA], [constants.DONE], adb_info, filesync_info):
             if cmd_id == constants.DONE:
                 break
 
-            dest.write(data)
+            await stream.write(data)
             if progress_callback:
                 progress.send(len(data))
 
-    async def push(self, source_file, device_filename, st_mode=constants.DEFAULT_PUSH_MODE, mtime=0, progress_callback=None, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S):
+    async def push(self, local_path, device_path, st_mode=constants.DEFAULT_PUSH_MODE, mtime=0, progress_callback=None, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S):
         """Push a file or directory to the device.
 
         Parameters
         ----------
-        source_file : str
-            Either a filename, a directory or file-like object to push to the device.
-        device_filename : str
-            Destination on the device to write to.
+        local_path : str
+            Either a filename or a directory to push to the device
+        device_path : str
+            Destination on the device to write to
         st_mode : int
-            Stat mode for filename
+            Stat mode for ``local_path``
         mtime : int
-            Modification time to set on the file.
+            Modification time to set on the file
         progress_callback : function, None
-            Callback method that accepts filename, bytes_written and total_bytes, total_bytes will be -1 for file-like
-            objects
+            Callback method that accepts ``device_path``, ``bytes_written``, and ``total_bytes``
         transport_timeout_s : float, None
-            Expected timeout for any part of the push.
+            Expected timeout for any part of the push
         read_timeout_s : float
             The total time in seconds to wait for a ``b'CLSE'`` or ``b'OKAY'`` command in :meth:`AdbDeviceAsync._read`
 
@@ -542,37 +539,36 @@ class AdbDeviceAsync(object):
         if not self.available:
             raise exceptions.AdbConnectionError("ADB command not sent because a connection to the device has not been established.  (Did you call `AdbDeviceAsync.connect()`?)")
 
-        if isinstance(source_file, str):
-            if os.path.isdir(source_file):
-                await self.shell("mkdir " + device_filename, transport_timeout_s, read_timeout_s)
-                for f in os.listdir(source_file):
-                    await self.push(os.path.join(source_file, f), device_filename + '/' + f, progress_callback=progress_callback)
-                return
+        local_path_is_dir, local_paths, device_paths = await get_running_loop().run_in_executor(None, get_files_to_push, local_path, device_path)
 
-        adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
-        filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_PUSH_FORMAT)
+        if local_path_is_dir:
+            await self.shell("mkdir " + device_path, transport_timeout_s, read_timeout_s)
 
-        with _open(source_file, 'rb') as source:
-            await self._open(b'sync:', adb_info)
-            await self._push(source, device_filename, st_mode, mtime, progress_callback, adb_info, filesync_info)
+        for _local_path, _device_path in zip(local_paths, device_paths):
+            adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
+            filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_PUSH_FORMAT, maxdata=self._maxdata)
 
-        await self._close(adb_info)
+            async with aiofiles.open(_local_path, 'rb') as stream:
+                await self._open(b'sync:', adb_info)
+                await self._push(stream, _device_path, st_mode, mtime, progress_callback, adb_info, filesync_info)
 
-    async def _push(self, datafile, filename, st_mode, mtime, progress_callback, adb_info, filesync_info):
+            await self._close(adb_info)
+
+    async def _push(self, stream, device_path, st_mode, mtime, progress_callback, adb_info, filesync_info):
         """Push a file-like object to the device.
 
         Parameters
         ----------
-        datafile : _io.BytesIO
+        stream : AsyncBufferedReader
             File-like object for reading from
-        filename : str
-            Filename to push to
+        device_path : str
+            Destination on the device to write to
         st_mode : int
-            Stat mode for filename
+            Stat mode for the file
         mtime : int
             Modification time
         progress_callback : function, None
-            Callback method that accepts ``filename``, ``bytes_written``, and ``total_bytes``
+            Callback method that accepts ``device_path``, ``bytes_written``, and ``total_bytes``
         adb_info : _AdbTransactionInfo
             Info and settings for this ADB transaction
 
@@ -582,17 +578,17 @@ class AdbDeviceAsync(object):
             Raised on push failure.
 
         """
-        fileinfo = ('{},{}'.format(filename, int(st_mode))).encode('utf-8')
+        fileinfo = ('{},{}'.format(device_path, int(st_mode))).encode('utf-8')
 
         await self._filesync_send(constants.SEND, adb_info, filesync_info, data=fileinfo)
 
         if progress_callback:
-            total_bytes = os.fstat(datafile.fileno()).st_size if isinstance(datafile, FILE_TYPES) else -1
-            progress = self._transport_progress(lambda current: progress_callback(filename, current, total_bytes))
+            total_bytes = (await get_running_loop().run_in_executor(os.fstat, stream.fileno())).st_size
+            progress = self._transport_progress(lambda current: progress_callback(device_path, current, total_bytes))
             next(progress)
 
         while True:
-            data = datafile.read(constants.MAX_PUSH_DATA)
+            data = await stream.read(self.max_chunk_size)
             if data:
                 await self._filesync_send(constants.DATA, adb_info, filesync_info, data=data)
 
@@ -612,12 +608,12 @@ class AdbDeviceAsync(object):
 
             raise exceptions.PushFailedError(data)
 
-    async def stat(self, device_filename, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S):
+    async def stat(self, device_path, transport_timeout_s=None, read_timeout_s=constants.DEFAULT_READ_TIMEOUT_S):
         """Get a file's ``stat()`` information.
 
         Parameters
         ----------
-        device_filename : str
+        device_path : str
             The file on the device for which we will get information.
         transport_timeout_s : float, None
             Expected timeout for any part of the pull.
@@ -640,8 +636,8 @@ class AdbDeviceAsync(object):
         adb_info = _AdbTransactionInfo(None, None, transport_timeout_s, read_timeout_s)
         await self._open(b'sync:', adb_info)
 
-        filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_STAT_FORMAT)
-        await self._filesync_send(constants.STAT, adb_info, filesync_info, data=device_filename)
+        filesync_info = _FileSyncTransactionInfo(constants.FILESYNC_STAT_FORMAT, maxdata=self._maxdata)
+        await self._filesync_send(constants.STAT, adb_info, filesync_info, data=device_path)
         _, (mode, size, mtime), _ = await self._filesync_read([constants.STAT], adb_info, filesync_info, read_data=False)
         await self._close(adb_info)
 
@@ -773,7 +769,7 @@ class AdbDeviceAsync(object):
             data = bytearray()
             while data_length > 0:
                 temp = await self._transport.bulk_read(data_length, adb_info.transport_timeout_s)
-                _LOGGER.debug("bulk_read(%d): %s", data_length, repr(temp))
+                _LOGGER.debug("bulk_read(%d): %.1000s", data_length, repr(temp))
 
                 data += temp
                 data_length -= len(temp)
@@ -1130,8 +1126,7 @@ class AdbDeviceAsync(object):
         Parameters
         ----------
         progress_callback : function
-            Callback method that accepts ``filename``, ``bytes_written``, and ``total_bytes``; total_bytes will be -1 for file-like
-            objects.
+            Callback method that accepts ``device_path``, ``bytes_written``, and ``total_bytes``
 
         """
         current = 0
